

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    Buffer description constants.                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


const int GET_BUF_SIZE          = 4 ;
const int TEG_BUF_SIZE          = 2 ;

const int GAR_BUF_SIZE          = 5 ;

const int PAR_BUF_SIZE          = 4 ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `GetServer' class.                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class GetServer : public Server {
  enum State {ALPHA_0, ALPHA_1, ALPHA_2} ;
  State state ;

  int dstId ;

  int getBuf [GET_BUF_SIZE] ;
  int tegBuf [TEG_BUF_SIZE] ;
  int datBufSize ;
  char* datBuf ;

  void unpackGetBuf(char**, int*, char**, int*) ;
  void packTegBuf(char*, int) ;
  void packDatBuf(int, char*, int) ;

  char* defData ;
public :
  GetServer(int, char*) ;
  ~GetServer() ;

  void transition(MPI_Status*) ;

  void getLoc(char*, int, char*, int) ;
} ;



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `GarServer' class.                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class GarServer : public Server {
  enum State {ALPHA_0, ALPHA_1, ALPHA_2, ALPHA_3, ALPHA_4} ;
  State state ;

  int blockSize, rank, dstId ;
  char *dstData, *srcData ;
  int *exts, *dstStrs, *srcStrs ;

  int garBuf [GAR_BUF_SIZE] ;

  int  gasBufSize ;
  int* gasBuf ;

  MPI_Datatype* arrayType ;

  void unpackGarBuf(int* buffer, int* blockSize, int* rank,
                    char** dstData, char** srcData, int* srcOffset) ;

  char* defData ;
public :
  GarServer(int, char*) ;
  ~GarServer() ;

  void transition(MPI_Status*) ;
} ;



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `ParServer' class.                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class ParServer : public Server {
  enum State {ALPHA_0, ALPHA_1, ALPHA_2, ALPHA_3} ;
  State state ;

  int blockSize, rank, srcId ;
  char *dstData, *srcData ;
  int *exts, *dstStrs, *srcStrs ;

  int parBuf [PAR_BUF_SIZE] ;

  int  pasBufSize ;
  int* pasBuf ;

  MPI_Datatype* arrayType ;

  void unpackParBuf(int* buffer, int* blockSize, int* rank,
                    char** dstData, int* dstOffset) ;

  char* defData ;
public :
  ParServer(int, char*) ;
  ~ParServer() ;

  void transition(MPI_Status*) ;
} ;



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `TegServer' class.                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class TegServer : public Server {
  int n ;
  enum State {GAMMA_0, GAMMA_1} ;
  State state ;

  int tegBuf [TEG_BUF_SIZE] ;
  int datBufSize ;
  char* datBuf ;

  void unpackTegBuf() ;
public :
  TegServer (int) ;
  ~TegServer () {}

  void transition(MPI_Status*) ;

  void notify() ;
  int active() ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `RagServer' class.                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class RagServer : public Server {
  int n ;
  enum State {GAMMA_0, GAMMA_1, GAMMA_2} ;
  State state ;

  int ragBuf [GAR_BUF_SIZE] ;
  int* sagBuf ;

  int blockSize, rank, srcId ;
  char *dstData ;
  int *exts, *dstStrs ;

  MPI_Datatype* arrayType ;

  void unpackRagBuf(int* buffer, int* blockSize, int* rank, char** dstData) ;
public :
  RagServer (int) ;
  ~RagServer () {}

  void transition(MPI_Status*) ;

  void notify() ;
  int active() ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `RapServer' class.                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class RapServer : public Server {
  int n ;

  int rapBuf [PAR_BUF_SIZE] ;
public :
  RapServer (int) ;
  ~RapServer () {}

  void transition(MPI_Status*) ;

  void notify() ;
  int active() ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `DataExchange' class.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //



class DataExchange {

  // Local servers
  GetServer* getServer ;
  TegServer* tegServer ;
  GarServer* garServer ;
  RagServer* ragServer ;
  ParServer* parServer ;
  RapServer* rapServer ;

  // State for client thread

  int locId ;
  char* defData ;
  ProcTree* spanTree ;

  int getBuf [GET_BUF_SIZE] ;
  void packGetBuf(char*, int, char*, int) ;
 
  void packGarBuf(int* garBuf, int blockSize, int rank,
		  char* dstData, char* srcData, int srcOffset) ;
  void packGasBuf(int* gasBuf, int rank,
                  int* exts, int* dstStrs, int* srcStrs) ;

  void packParBuf(int* parBuf, int blockSize, int rank,
		  char* dstData, int dstOffset) ;
  void packPasBuf(int* pasBuf, int rank, int* exts, int* dstStrs) ;

  void arrcpy(int blockSize, int rank, int* exts,
	      char* dstData, int* dstStrs, char* srcData, int* srcStrs) ;
public :
  DataExchange(ProcTree* spanTree, char* defData) ;
  ~DataExchange() ;

  void get(char* locData, int blockSize,
           int remId, char* remData, int remOffset) ;
  void put(char* locData, int blockSize,
           int remId, char* remData, int remOffset) ;

  void getArray(int blockSize, int rank, int* exts,
	        char* dstData, int* dstStrs,
	        int srcId, char* srcData, int srcOffset, int* srcStrs) ;

  void putArray(int blockSize, int rank, int* exts,
	        char* srcData, int* srcStrs,
	        int dstId, char* dstData, int dstOffset, int* dstStrs) ;
} ;


